#include <linux/kernel.h>
#include <linux/platform_device.h>
#include <asm/avm_enh/avm_mtd.h>

static int __init __prepare_ram_resource(char *p, char *name,
					int group_selector)
{
	char *np;
	unsigned long start, end;
	int result = 0;

	pr_err("[%s] entered\n", __func__);

	if (p == NULL) {
		result = -EINVAL;
		goto err_out;
	}

	/* parse start and end addresses */
	np = strchr(p, ',');
	if (np == NULL) {
		result = -EINVAL;
		goto err_out;
	}

	*np = '\0';
	++np;

	result = kstrtoul(p, 0, &start);
	if(result != 0) {
		goto err_out;
	}

	result = kstrtoul(np, 0, &end);
	if(result != 0) {
		goto err_out;
	}

	if (start >= end) {
		result = -ENOSPC;
		goto err_out;
	}

	result = avm_mtd_set_mtdram_info(name, group_selector, start, end,
					IORESOURCE_MEM);
	if (result != 0)
		goto err_out;

	pr_info("[%s] mtdram 0x%08lx-0x%08lx\n", __func__, start, end);

	return result;

err_out:
	pr_err("[%s] error parsing setup string: %s\n", __func__, p);
	return result;
}

static int __init prepare_ram_resource0(char *p)
{
	return __prepare_ram_resource(p, AVM_DEFAULT_MTD_NAME, GROUP_MTDS);
}

static int __init prepare_ram_resource_n(char *p)
{
	char *name;

	/* parse device name */
	name = p;
	strsep(&p, ",");

	if (!p)
		return -EINVAL;

	return __prepare_ram_resource(p, name, GROUP_FILESYSTEMS);
}

early_param("mtdram1", prepare_ram_resource0);
early_param("mtdram", prepare_ram_resource_n);
